package meta.system.monitors;

import java.io.PrintWriter;
import java.io.StringWriter;

/**
 * Adds an interested collection to UpDown implementors and provides logic for informing them of the state of the resource. 
 */
public aspect UpDownMonitorAugment extends MonitorAugment<UpDownMonitorInterest>{

	boolean UpDownMonitor.resourceCurrentlyAvailible = true;
		
	public void UpDownMonitor.run(){
		Thread.currentThread().setName(this.getClass().getName());
		boolean currentState = false;
		
		try{
			currentState = this.resourceAvailible();
		}catch(Exception e){
			StringWriter stackTrace = new StringWriter();
			e.printStackTrace(new PrintWriter(stackTrace));			
			log.fatal(e+" occured when checking for resource availibillity. Assuming its unavailible. Trace - "+stackTrace);
		}
		
		if(currentState && !resourceCurrentlyAvailible){
			log.debug("Resource has become availible");
			//then the resource has become availbe where previously it was not
			informInterestedOfUp();
			//store state
			resourceCurrentlyAvailible = true;
		}else{
			if(!currentState && resourceCurrentlyAvailible){
				log.debug("Resource has become unavailible");
				//then the resource has become unavailible
				informInterestedOfDown();
				//store state
				resourceCurrentlyAvailible = false;
			}
		}
		//log.debug("Updating run at for "+this.getClass().getName());
		this.lastRun = System.currentTimeMillis();
		
		this.pool.execute(this);
	}

	public void UpDownMonitor.informInterestedOfUp(){
		for(UpDownMonitorInterest interest : interested){
			try{
				interest.resourceAvailible(this);
			}catch(Exception e){
				StringWriter stackTrace = new StringWriter();
				e.printStackTrace(new PrintWriter(stackTrace));	
				log.warn("Exception occured when trying to inform "+interest.getClass().getName()+" of a resource returning to availible. Trace -\n"+stackTrace);				
			}
		}
	}
	
	public void UpDownMonitor.informInterestedOfDown(){
		for(UpDownMonitorInterest interest : interested){
			try{
				interest.resourceUnavailible(this);
			}catch(Exception e){
				StringWriter stackTrace = new StringWriter();
				e.printStackTrace(new PrintWriter(stackTrace));	
				log.warn("Exception occured when trying to inform "+interest.getClass().getName()+" of a resource failiure. Trace -\n"+stackTrace);
			}
		}		
	}

	public void UpDownMonitor.addInterested(UpDownMonitorInterest inter){
		this.interested.add(inter);
	}
	public void UpDownMonitor.removeInterested(UpDownMonitorInterest inter){
		this.interested.remove(inter);
	}
}
