package meta.system.registry;

/**
 * Base Interface for entitys wishing to be stored in the registry system
 * @author tom
 */
public interface Registable {
	public String getName();
	public void setName(String name);
	
	/**
	 * If your class is unable to set a name on construction, implement this and provide one. Will only be called
	 * on the first registration of the entity. Should be used by all none {@link meta.services.MetaService} (which
	 * should pass their static name in the constructor) to generate a name from their contents or by other means.
	 * @return
	 */
	public String provideName();	
	public void setParent(Registry parent);
	public Registry getParent();
	
	
	/*
	 * Registry life cycle methods.
	 */
	
	/**
	 * Called just before entity is registered. Any exceptions being thrown from this method will result in the entity
	 * not being stored in the registry. 
	 */
	public void registering() throws Exception;
	/**
	 * Called after the entity has been placed in the hash. Exceptions occuring here will cause the entity to be removed from
	 * the registry in question and a RegistrationException is thrown.
	 */
	public void registered() throws Exception;
	
	/**
	 * Called just before the entity is deregistered. Any exceptions thrown from this method will result in the entity not
	 * being removed (in the case of cacheables their life time will be set null making them immortal) and a RegistrationException
	 * will be thrown. Caches provide a 'forceDrop' variable to override this concept and it is not implemented in the purge/flush methods.
	 */
	public void deregistering() throws Exception;
	/**
	 * Called after the entity has been removed from the registry. Any exceptions thrown here will be passed up, but the entity
	 * will still be removed from the registry.
	 */
	public void deregistered() throws Exception;
}
