package meta.system.xml;


import java.util.*;
import java.lang.reflect.*;
import javolution.util.*;
import meta.utils.xml.XMLExtractor;

/**
 * Adds XML conversion logic to XMLRepresnetable implementors.
 */
public aspect XMLRepresentableAugment {

	static Map<String, Collection<Field>> xmlFields = new FastMap<String, Collection<Field>>();
	
	/*
	 * XML creation methods
	 */
	
	public String XMLRepresentable.createXMLRep(){
		return genericCreateXMLRep(this);
	}
	
	private static String genericCreateXMLRep(XMLRepresentable instance){
		
		Collection<Field> classFields = null;
		
		String className = instance.getClass().getName();
		
		if((classFields= xmlFields.get(className)) == null){
			//then its the first time one of these has been converted
			classFields = getAllClassXMLFields(instance.getClass());
			xmlFields.put(className, classFields);
		}
		
		StringBuilder billy = new StringBuilder();//TODO use pool?
		
		billy.append('<'+className+'>');
		
		//now add all the fields
		try{
			XMLRepresentable assignedIn = null; 
			for(Field filly : classFields){
				billy.append('<'+filly.getName()+'>');
				//is an XMLRepresentable entity in its own right?
				if(isXMLRepresentable(filly, instance)){
					//test for null 
					assignedIn = ((XMLRepresentable)filly.get(instance));
					if(assignedIn != null){
						//gets its representation
						billy.append(assignedIn.createXMLRep());
					}else{
						billy.append("null");
					}
					billy.append("</"+filly.getName()+'>');	
					continue;
				}
				//is it a collection of XMLRepresentables?
				if(Collection.class.isAssignableFrom(filly.getType())){
					createCollectionXML(instance, filly, billy);
					billy.append("</"+filly.getName()+'>');	
					continue;
				}
				
				if(filly.getType().isArray()){
					createArrayXML(instance, filly, billy);
					billy.append("</"+filly.getName()+'>');	
					continue;
				}
				//specialtypetastic

				if(filly.getType().isEnum()){
					billy.append('<'+filly.getType().getName()+'>');
					billy.append(((Enum)filly.get(instance)).ordinal());
					billy.append("</"+filly.getType().getName()+'>');					
					billy.append("</"+filly.getName()+'>');	
					continue;
				}
				
				//its yr run of the mill field
				billy.append(filly.get(instance));
				
				billy.append("</"+filly.getName()+'>');			
			}
		}catch(Exception e){
			throw new IllegalStateException(e+" occured when trying to create an XML rep of "+className+".", e);			
		}
		billy.append("</"+className+'>');	
		
		return billy.toString();
	}
	
	private static void createCollectionXML(XMLRepresentable instance, Field filly, StringBuilder billy) throws IllegalArgumentException, IllegalAccessException{
		Collection<XMLRepresentable> coll = (Collection<XMLRepresentable>)filly.get(instance);
		
		if(coll == null){
			billy.append("null");
			return;		
		}
		
		//add coll type tags
		billy.append('<'+coll.getClass().getName()+'>');
		
		for(XMLRepresentable inst : coll){
			billy.append(inst.createXMLRep());
		}
		billy.append("</"+coll.getClass().getName()+'>');		
	}
	
	private static void createArrayXML(XMLRepresentable instance, Field filly, StringBuilder billy) throws IllegalArgumentException, IllegalAccessException {
		
		Object[] harry = (Object[])filly.get(instance);
		
		if(harry == null){
			billy.append("null");
			return;		
		}
		
		XMLRepresentable harriet = null;String className = null;
		//discover true class name

		for(int i=0;i<harry.length;i++){
			harriet = ((XMLRepresentable)harry[i]);
			if(harriet != null){
				className = harriet.getClass().getName();
				break;
			}			
		}
		
		if(className == null){
			//then this array only contains nulls
			billy.append("null");
			return;
		}
		
		for(int i=0;i<harry.length;i++){
			harriet = ((XMLRepresentable)harry[i]);
			if(harriet == null){
				billy.append('<'+className+">null</"+className+'>');
			}else{
				billy.append(harriet.createXMLRep());
			}
		}		
	}
	
	
	
	/*
	 * Set from XML string
	 */
	
	public void XMLRepresentable.setFromXML(String xml){
		genericSetFromXML(this, xml);
	}
	
	private static void genericSetFromXML(XMLRepresentable instance, String xml){
		if(xml == null || xml.equalsIgnoreCase("null"))return; //this instance has no settings
		Collection<Field> classFields = null;

		String className = instance.getClass().getName();
		
		if((classFields= xmlFields.get(className)) == null){
			//then its the first time one of these has been converted
			classFields = getAllClassXMLFields(instance.getClass());
			xmlFields.put(className, classFields);
		}
		
		//strip the class outer tags
		xml = XMLExtractor.stripContainingTagPair(xml);
		if(xml.equalsIgnoreCase("null")){
			//then instance had no state worth serialzing on convertion
			return;
		}
		
		try{
			String fillyXml = null;
			for(Field filly : classFields){
				fillyXml = XMLExtractor.genericXMLContentExtracter(xml, filly.getName(), filly.getName());
				
				//crop off this filly's xml
				xml = xml.substring(fillyXml.length());
				
				fillyXml = XMLExtractor.stripContainingTagPair(fillyXml);	
				if(fillyXml == null || fillyXml.equalsIgnoreCase("null") ){
					//filly.set(this, null);
					continue;
				}
				if(isXMLRepresentable(filly, instance)){
					if(fillyXml == null || fillyXml.equalsIgnoreCase("null"))continue;
					Object obj = filly.get(instance);
					if(obj == null){
						//instanciate one :)
						String name = fillyXml.substring(1, fillyXml.indexOf('>'));
						obj = Class.forName(name).newInstance();					
						//throw new IllegalStateException(this.getClass().getName()+" contains an XMLRepresentable field that is not initized, it must be.");
					}
					((XMLRepresentable)obj).setFromXML(fillyXml);
					filly.set(instance, obj);
					continue;
				}
				//is it a collection of XMLRepresentables?
				if(Collection.class.isAssignableFrom(filly.getType())){
					filly.set(instance, createCollectionObject(fillyXml, filly));
					continue;
				}
				//is it an array field?
				if(filly.getType().isArray()){
					filly.set(instance, createArrayObject(fillyXml, filly));
					continue;
				}
				
				if(filly.getType().isEnum()){
					filly.set(instance, createEnumObject(fillyXml, filly));
					continue;					
				}
				
				//its yr run of the mill field
				filly.set(instance, parseToObject(filly, fillyXml));											
			}
		}catch(Exception e){
			throw new IllegalStateException(e+" occured when trying to set a "+className+" from "+xml+".", e);
		}		
	}
	
	
	private static Collection createCollectionObject(String collXml, Field filly) throws InstantiationException, IllegalAccessException, ClassNotFoundException{
		if(collXml == null || collXml.equalsIgnoreCase("null"))return null;//the collection did'nt have any elements
		
		String collType = collXml.substring(1, collXml.indexOf('>'));
		
		Collection collin = (Collection)Class.forName(collType).newInstance(); 
		
		//this method prohibits multi type collections, if u want them have to Class.forName for each member of coll
		
		int sndAngle = collXml.indexOf('<', collXml.indexOf('>'));
		
		String elleType = collXml.substring(sndAngle+1, collXml.indexOf('>', sndAngle));
		
		String elleOpen = '<'+elleType+'>';String elleClose = "</"+elleType+'>';
		String indiElle = null;XMLRepresentable nelle = null;
		for(int i=1;;i++){
			indiElle = XMLExtractor.genericXMLContentExtracter(collXml, elleOpen, elleClose, i);
			if(indiElle == null){
				break;
			}
			if(XMLExtractor.stripContainingTagPair(indiElle).equalsIgnoreCase("null")){
				collin.add(null);
				continue;
			}
			
			nelle = (XMLRepresentable)Class.forName(elleType).newInstance();
			nelle.setFromXML(indiElle);
			collin.add(nelle);
		}
		
		return collin;
	}
	
	private static Object[] createArrayObject(String arrXml, Field filly) throws InstantiationException, IllegalAccessException, ClassNotFoundException{
		Collection tempCollin = createCollectionObject("<java.util.LinkedList>"+arrXml+"</java.util.LinkedList>", filly);
		
		//convert into premative array

		String elleType = arrXml.substring(1, arrXml.indexOf('>'));
		Class elleClass = Class.forName(elleType);
		
		Object[] harry = (Object[])Array.newInstance(elleClass, tempCollin.size());
		
		tempCollin.toArray(harry);
		
		return harry;
	}
	
	
	private static Enum createEnumObject(String enumXml, Field filly) throws InstantiationException, IllegalAccessException, ClassNotFoundException{

		String enumType = enumXml.substring(1, enumXml.indexOf('>'));
		
		String value = XMLExtractor.extractFieldValue(enumXml, enumType);
		if(value == null)return null;
		
		Class enumClass = Class.forName(enumType);
		
		//get the sent ordinal value
		int sentOrd = Integer.parseInt(value);
		
		EnumSet<? extends Enum> enumerSet = EnumSet.allOf(enumClass);
		
		for(Enum oneOf : enumerSet){
			if(oneOf.ordinal() == sentOrd)return oneOf;
		}
		throw new IllegalStateException("The sent enum ordinal "+sentOrd+" did not match any ordinal of "+enumType+". Please check versions of this" +
				" enum are the same in the location this XML was created and where it should have been reconstructed.");
	}
		
	/*
	 * Util Methods
	 */
	
	public static boolean isXMLRepresentable(Field filly, Object that) throws IllegalArgumentException, IllegalAccessException{
		if(XMLRepresentable.class.isAssignableFrom(filly.getType())){
			return true;
		}
		return false;
	}
	
	
	public static Collection<Field> getAllClassXMLFields(Class theClass){
		
		Collection<Field> xmlFields = new LinkedList<Field>();
		
		Field[] harry = theClass.getDeclaredFields();
		
		for(Field filly : harry){
			if(filly.getAnnotation(IncludeInXML.class) != null){
				filly.setAccessible(true);
				xmlFields.add(filly);
			}
		}
		
		//get all super deffined xml fields as well
		if(theClass.getSuperclass() != null)
		xmlFields.addAll(getAllClassXMLFields(theClass.getSuperclass()));
		
		return xmlFields;		
	}
	
	/**
	 * All nifty things have to have abit of donkey in them, here it is. Parses the passed String into the
	 * object type of the passed field.
	 */
	public static Object parseToObject(Field filly, String xml){
		
		//String value = XMLExtractor.extractFieldValue(xml, filly.getName());
		String value = xml;
		
		if(value == null || value.equalsIgnoreCase("null"))return null;
		
		Class fillyClass = filly.getType();
		
		//Yeah this couold be done with abit more reflection or including the field type in the xml, but this works and is simple dimple
		if(fillyClass.isAssignableFrom(String.class)){
			return value;
		}
		if(fillyClass.getName().toLowerCase().contains("integer") || fillyClass.getName().toLowerCase().contains("int")){
			Integer var = new Integer(value);
			if(fillyClass.isPrimitive()){
				return var.intValue();
			}else{
				return var;
			}
		}		
		if(fillyClass.getName().toLowerCase().contains("boolean")){
			Boolean var = new Boolean(value);
			if(fillyClass.isPrimitive()){
				return var.booleanValue();
			}else{
				return var;
			}
		}
		
		if(fillyClass.getName().toLowerCase().contains("double")){
			Double var = new Double(value);
			if(fillyClass.isPrimitive()){
				return var.doubleValue();
			}else{
				return var;
			}
		}
		if(fillyClass.getName().toLowerCase().contains("long")){
			Long var = new Long(value);
			if(fillyClass.isPrimitive()){
				return var.longValue();
			}else{
				return var;
			}
		}
		if(fillyClass.getName().toLowerCase().contains("float")){
			Float var = new Float(value);
			if(fillyClass.isPrimitive()){
				return var.floatValue();
			}else{
				return var;
			}
		}
		if(fillyClass.getName().toLowerCase().contains("short")){
			Short var = new Short(value);
			if(fillyClass.isPrimitive()){
				return var.shortValue();
			}else{
				return var;
			}
		}
		
		throw new IllegalStateException("Unable to turn "+value+" into an instance of "+fillyClass+". Please add this type to the parseObject method.");
	}
	
}
